import itertools
import os
import sys
import uuid
from datetime import date
from pathlib import Path, PurePosixPath, PureWindowsPath

import pytest

from ophyd_async.core import (
    AutoIncrementFilenameProvider,
    AutoIncrementingPathProvider,
    StaticPathProvider,
    UUIDFilenameProvider,
    YMDPathProvider,
)


def test_auto_increment_filename_provider(static_path_provider_factory):
    auto_inc_fp = AutoIncrementFilenameProvider(inc_delimeter="")
    dp = static_path_provider_factory(auto_inc_fp)

    # Our filenames should go from 00000 to 99999.
    # We increment by one each time, so just check if casting filename to
    # int gets us i.
    for i in range(100000):
        info = dp()
        assert int(info.filename) == i

    # Since default max digits is 5, incrementing one more time to 100000
    # will go over the limit and raise a value error
    with pytest.raises(ValueError):
        dp()


@pytest.mark.parametrize(
    "uuid_version", [uuid.uuid1, uuid.uuid3, uuid.uuid4, uuid.uuid5]
)
def test_uuid_filename_provider(static_path_provider_factory, uuid_version):
    uuid_call_args = []
    if uuid_version in [uuid.uuid3, uuid.uuid5]:
        uuid_call_args = [uuid.NAMESPACE_URL, "test"]
    uuid_fp = UUIDFilenameProvider(
        uuid_call_func=uuid_version, uuid_call_args=uuid_call_args
    )

    dp = static_path_provider_factory(uuid_fp)

    info = dp()

    # Try creating a UUID object w/ given version to check if it is valid uuid
    uuid.UUID(info.filename, version=int(uuid_version.__name__[-1]))


@pytest.mark.parametrize("uuid_version", [uuid.uuid3, uuid.uuid5])
def test_uuid_filename_provider_no_namespace(
    static_path_provider_factory, uuid_version
):
    uuid_fp = UUIDFilenameProvider(uuid_call_func=uuid_version)
    dp = static_path_provider_factory(uuid_fp)

    with pytest.raises(ValueError):
        dp()


def test_auto_increment_path_provider(static_filename_provider, tmp_path):
    auto_inc_path_provider = AutoIncrementingPathProvider(
        static_filename_provider, tmp_path, num_calls_per_inc=3, increment=2
    )

    for _ in range(3):
        info = auto_inc_path_provider()
        assert os.path.basename(info.directory_path) == "00000"
    info = auto_inc_path_provider()
    assert os.path.basename(info.directory_path) == "00002"


def test_ymd_path_provider(static_filename_provider, tmp_path):
    ymd_path_provider = YMDPathProvider(static_filename_provider, tmp_path)
    current_date = date.today()
    date_path = (
        f"{current_date.year:04d}/{current_date.month:02d}/{current_date.day:02d}"
    )

    info_a = ymd_path_provider()
    assert info_a.directory_path == tmp_path / date_path

    info_b = ymd_path_provider(device_name="test_device")
    assert info_b.directory_path == tmp_path / "test_device" / date_path


def test_path_provider_fails_non_absolute_path(static_filename_provider):
    """Test that the path provider raises an error for non-absolute paths."""

    path_provider = StaticPathProvider(
        static_filename_provider, Path("non_absolute_path")
    )

    with pytest.raises(ValueError, match="directory_path must be an absolute path"):
        path_provider()


PATH_FORMATS = [
    PurePosixPath,
    PureWindowsPath,
]


@pytest.mark.parametrize(
    "write_path_format, read_path_format",
    list(itertools.product(PATH_FORMATS, PATH_FORMATS)),
)
def test_path_provider_with_different_uri(
    write_path_format, read_path_format, static_filename_provider
):
    """Test that the path provider can handle a custom URI."""

    posix_path_a = PurePosixPath("/tmp/posix_path_a")
    posix_path_b = PurePosixPath("/tmp/posix_path_b")

    windows_base_path = "C:\\Users\\test\\AppData\\Local\\Temp"
    windows_path_a = PureWindowsPath(f"{windows_base_path}\\windows_path_a")
    windows_path_b = PureWindowsPath(f"{windows_base_path}\\windows_path_b")

    write_path = posix_path_a
    if write_path_format is PureWindowsPath:
        write_path = windows_path_a

    read_path = posix_path_b
    if read_path_format is PureWindowsPath:
        read_path = windows_path_b

    path_provider = StaticPathProvider(
        static_filename_provider,
        write_path,
        directory_uri=f"file://localhost/{read_path.as_posix().lstrip('/')}/",
    )

    info = path_provider()

    if write_path_format is PureWindowsPath:
        assert str(info.directory_path) == f"{windows_base_path}\\windows_path_a"
    else:
        assert str(info.directory_path) == "/tmp/posix_path_a"

    if read_path_format is PureWindowsPath:
        assert (
            info.directory_uri
            == "file://localhost/C:/Users/test/AppData/Local/Temp/windows_path_b/"
        )
    else:
        assert info.directory_uri == "file://localhost/tmp/posix_path_b/"


@pytest.mark.skipif(sys.platform == "win32", reason="Unix specific test")
def test_windows_path_produces_valid_path_info_on_unix(static_filename_provider):
    path = PureWindowsPath("C:\\Users\\test\\AppData\\Local\\Temp\\windows_path")
    path_provider = StaticPathProvider(static_filename_provider, path)
    info = path_provider()
    assert (
        str(info.directory_path)
        == "C:\\Users\\test\\AppData\\Local\\Temp\\windows_path"
    )
    assert (
        info.directory_uri
        == "file://localhost/C:/Users/test/AppData/Local/Temp/windows_path/"
    )


@pytest.mark.skipif(sys.platform != "win32", reason="Windows specific test")
def test_posix_path_produces_valid_path_info_on_windows(static_filename_provider):
    path = PurePosixPath("/tmp/posix_path")
    path_provider = StaticPathProvider(static_filename_provider, path)
    info = path_provider()
    assert str(info.directory_path) == "/tmp/posix_path"
    assert info.directory_uri == "file://localhost/tmp/posix_path/"



import asyncio
from typing import Generic, TypeVar

from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    AsyncStatus,
    DetectorController,
    DetectorTrigger,
    EnableDisable,
    TriggerInfo,
    observe_value,
    set_and_wait_for_value,
)
from ophyd_async.epics.core import stop_busy_record

from ._core_io import (
    ADBaseIO,
    ADState,
    NDCBFlushOnSoftTrgMode,
    NDPluginCBIO,
)
from ._utils import ADImageMode

# Default set of states that we should consider "good" i.e. the acquisition
#  is complete and went well
DEFAULT_GOOD_STATES: frozenset[ADState] = frozenset([ADState.IDLE, ADState.ABORTED])

ADBaseIOT = TypeVar("ADBaseIOT", bound=ADBaseIO)
ADBaseControllerT = TypeVar("ADBaseControllerT", bound="ADBaseController")


class ADBaseController(DetectorController, Generic[ADBaseIOT]):
    def __init__(
        self,
        driver: ADBaseIOT,
        good_states: frozenset[ADState] = DEFAULT_GOOD_STATES,
        image_mode: ADImageMode = ADImageMode.MULTIPLE,
    ) -> None:
        self.driver: ADBaseIOT = driver
        self.good_states = good_states
        self.image_mode = image_mode
        self.frame_timeout = DEFAULT_TIMEOUT
        self._arm_status: AsyncStatus | None = None

    async def prepare(self, trigger_info: TriggerInfo) -> None:
        if trigger_info.trigger != DetectorTrigger.INTERNAL:
            msg = (
                "fly scanning (i.e. external triggering) is not supported for this "
                "device"
            )
            raise TypeError(msg)
        self.frame_timeout = (
            DEFAULT_TIMEOUT + await self.driver.acquire_time.get_value()
        )
        await asyncio.gather(
            self.driver.num_images.set(trigger_info.total_number_of_exposures),
            self.driver.image_mode.set(self.image_mode),
        )

    async def arm(self):
        self._arm_status = await self.start_acquiring_driver_and_ensure_status()

    async def wait_for_idle(self):
        if self._arm_status and not self._arm_status.done:
            await self._arm_status
        self._arm_status = None

    async def disarm(self):
        # We can't use caput callback as we already used it in arm() and we can't have
        # 2 or they will deadlock
        await stop_busy_record(self.driver.acquire, False)

    async def set_exposure_time_and_acquire_period_if_supplied(
        self,
        exposure: float | None = None,
        timeout: float = DEFAULT_TIMEOUT,
    ) -> None:
        """Set the exposure time and acquire period.

        If exposure is not None, this sets the acquire time to the exposure time
        and sets the acquire period to the exposure time plus the deadtime. This
        is expected behavior for most AreaDetectors, but some may require more
        specialized handling.

        :param exposure: Desired exposure time, this is a noop if it is None.
        :type exposure: How long to wait for the exposure time and acquire
            period to be set.
        """
        if exposure is not None:
            full_frame_time = exposure + self.get_deadtime(exposure)
            await asyncio.gather(
                self.driver.acquire_time.set(exposure, timeout=timeout),
                self.driver.acquire_period.set(full_frame_time, timeout=timeout),
            )

    async def start_acquiring_driver_and_ensure_status(
        self,
        start_timeout: float = DEFAULT_TIMEOUT,
        state_timeout: float = DEFAULT_TIMEOUT,
    ) -> AsyncStatus:
        """Start acquiring driver, raising ValueError if the detector is in a bad state.

        This sets driver.acquire to True, and waits for it to be True up to a timeout.
        Then, it checks that the DetectorState PV is in DEFAULT_GOOD_STATES,
        and otherwise raises a ValueError.


        :param start_timeout:
            Timeout used for waiting for the driver to start
            acquiring.
        :param state_timeout:
            Timeout used for waiting for the detector to be in a good
            state after it stops acquiring.
        :returns AsyncStatus:
            An AsyncStatus that can be awaited to set driver.acquire to True and perform
            subsequent raising (if applicable) due to detector state.

        """
        status = await set_and_wait_for_value(
            self.driver.acquire,
            True,
            timeout=start_timeout,
            wait_for_set_completion=False,
        )

        async def complete_acquisition() -> None:
            await status
            state = None
            try:
                async for state in observe_value(
                    self.driver.detector_state, done_timeout=state_timeout
                ):
                    if state in self.good_states:
                        return
            except TimeoutError as exc:
                if state is not None:
                    raise ValueError(
                        f"Final detector state {state.value} not in valid end "
                        f"states: {self.good_states}"
                    ) from exc
                else:
                    # No updates from the detector, something else is wrong
                    raise TimeoutError(
                        "Could not monitor detector state: "
                        + self.driver.detector_state.source
                    ) from exc

        return AsyncStatus(complete_acquisition())


class ADBaseContAcqController(ADBaseController[ADBaseIO]):
    """Continuous acquisition interface for an AreaDetector."""

    def __init__(self, driver: ADBaseIO, cb_plugin: NDPluginCBIO) -> None:
        self.cb_plugin = cb_plugin
        super().__init__(driver)

    def get_deadtime(self, exposure):
        # For now just set this to something until we can decide how to pass this in
        return 0.001

    async def ensure_acquisition_settings_valid(
        self, trigger_info: TriggerInfo
    ) -> None:
        """Ensure the trigger mode is valid for the detector."""
        if trigger_info.trigger != DetectorTrigger.INTERNAL:
            # Note that not all detectors will use the `DetectorTrigger` enum
            raise TypeError(
                "The continuous acq interface only supports internal triggering."
            )

        # Not all detectors allow for changing exposure times during an acquisition,
        # so in this least-common-denominator implementation check to see if
        # exposure time matches the current exposure time.
        exposure_time = await self.driver.acquire_time.get_value()
        if trigger_info.livetime is not None and trigger_info.livetime != exposure_time:
            raise ValueError(
                f"Detector exposure time currently set to {exposure_time}, "
                f"but requested exposure is {trigger_info.livetime}"
            )

    async def ensure_in_continuous_acquisition_mode(self) -> None:
        """Ensure the detector is in continuous acquisition mode."""
        image_mode = await self.driver.image_mode.get_value()
        acquiring = await self.driver.acquire.get_value()

        if image_mode != ADImageMode.CONTINUOUS or not acquiring:
            raise RuntimeError(
                "Driver must be acquiring in continuous mode to use the "
                "cont acq interface"
            )

    async def prepare(self, trigger_info: TriggerInfo) -> None:
        # These are broken out into seperate functions to make it easier to
        # override them in subclasses, for example if you want `prepare` to
        # setup the detector in continuous mode if it isn't already (for now,
        # we assume it already is). If your detector uses different enums
        # for `ImageMode` or `DetectorTrigger`, you should also override these.
        await self.ensure_acquisition_settings_valid(trigger_info)
        await self.ensure_in_continuous_acquisition_mode()

        # Configure the CB plugin to collect the correct number of triggers
        await asyncio.gather(
            self.cb_plugin.enable_callbacks.set(EnableDisable.ENABLE),
            self.cb_plugin.pre_count.set(0),
            self.cb_plugin.post_count.set(trigger_info.total_number_of_exposures),
            self.cb_plugin.preset_trigger_count.set(1),
            self.cb_plugin.flush_on_soft_trg.set(NDCBFlushOnSoftTrgMode.ON_NEW_IMAGE),
        )

    async def arm(self) -> None:
        # Start the CB plugin's capture, and cache it in the arm status attr
        self._arm_status = await set_and_wait_for_value(
            self.cb_plugin.capture,
            True,
            timeout=DEFAULT_TIMEOUT,
            wait_for_set_completion=False,
        )

        # Send the trigger to begin acquisition
        await self.cb_plugin.trigger.set(True, wait=False)

    async def disarm(self) -> None:
        await stop_busy_record(self.cb_plugin.capture, False)
        if self._arm_status and not self._arm_status.done:
            await self._arm_status
        self._arm_status = None



import asyncio
from collections.abc import Sequence

from bluesky.protocols import Triggerable

from ophyd_async.core import AsyncStatus, SignalR, StandardReadable
from ophyd_async.core import StandardReadableFormat as Format

from ._core_io import ADBaseIO, NDPluginBaseIO
from ._utils import ADImageMode


class SingleTriggerDetector(StandardReadable, Triggerable):
    def __init__(
        self,
        drv: ADBaseIO,
        read_uncached: Sequence[SignalR] = (),
        name="",
        plugins: dict[str, NDPluginBaseIO] | None = None,
    ) -> None:
        self.drv = drv
        if plugins is not None:
            for k, v in plugins.items():
                setattr(self, k, v)

        self.add_readables(
            [self.drv.array_counter, *read_uncached],
            Format.HINTED_UNCACHED_SIGNAL,
        )

        self.add_readables([self.drv.acquire_time], Format.CONFIG_SIGNAL)

        super().__init__(name=name)

    @AsyncStatus.wrap
    async def stage(self) -> None:
        await asyncio.gather(
            self.drv.image_mode.set(ADImageMode.SINGLE),
            self.drv.wait_for_plugins.set(True),
        )
        await super().stage()

    @AsyncStatus.wrap
    async def trigger(self) -> None:
        await self.drv.acquire.set(True)



import logging
import os
import sys
import typing
from collections.abc import Mapping, Sequence
from functools import cache
from math import isnan, nan
from typing import Any, Generic, cast

import numpy as np
from aioca import (
    FORMAT_CTRL,
    FORMAT_RAW,
    FORMAT_TIME,
    CANothing,
    Subscription,
    caget,
    camonitor,
    caput,
)
from aioca.types import AugmentedValue, Dbr, Format
from bluesky.protocols import Reading
from epicscorelibs.ca import dbr
from event_model import DataKey, Limits, LimitsRange

from ophyd_async.core import (
    Array1D,
    Callback,
    NotConnectedError,
    SignalDatatype,
    SignalDatatypeT,
    SignalMetadata,
    get_enum_cls,
    get_unique,
    make_datakey,
    wait_for_connection,
)

from ._util import EpicsSignalBackend, format_datatype, get_supported_values

logger = logging.getLogger("ophyd_async")


def _all_updates() -> bool:
    return os.environ.get("OPHYD_ASYNC_EPICS_CA_KEEP_ALL_UPDATES", "True") == "True"


def _limits_from_augmented_value(value: AugmentedValue) -> Limits:
    def get_limits(limit: str) -> LimitsRange | None:
        low = getattr(value, f"lower_{limit}_limit", nan)
        high = getattr(value, f"upper_{limit}_limit", nan)
        if not (isnan(low) and isnan(high)) and not high == low == 0:
            return LimitsRange(
                low=None if isnan(low) else low,
                high=None if isnan(high) else high,
            )
        return None

    limits = Limits()
    if limits_range := get_limits("alarm"):
        limits["alarm"] = limits_range
    if limits_range := get_limits("ctrl"):
        limits["control"] = limits_range
    if limits_range := get_limits("disp"):
        limits["display"] = limits_range
    if limits_range := get_limits("warning"):
        limits["warning"] = limits_range
    return limits


def _metadata_from_augmented_value(
    datatype: type[SignalDatatypeT] | None,
    value: AugmentedValue,
    metadata: SignalMetadata,
) -> SignalMetadata:
    metadata = metadata.copy()
    if hasattr(value, "units") and datatype not in (str, bool):
        metadata["units"] = value.units
    if (
        hasattr(value, "precision")
        and not isnan(value.precision)
        and datatype is not int
    ):
        metadata["precision"] = value.precision
    if (limits := _limits_from_augmented_value(value)) and datatype is not bool:
        metadata["limits"] = limits
    return metadata


class CaConverter(Generic[SignalDatatypeT]):
    def __init__(
        self,
        datatype: type[SignalDatatypeT],
        read_dbr: Dbr,
        write_dbr: Dbr | None = None,
        metadata: SignalMetadata | None = None,
    ):
        self.datatype = datatype
        self.read_dbr: Dbr = read_dbr
        self.write_dbr: Dbr | None = write_dbr
        self.metadata = metadata or SignalMetadata()

    def write_value(self, value: Any) -> Any:
        # The ca library will do the conversion for us
        return value

    def value(self, value: AugmentedValue) -> SignalDatatypeT:
        # for channel access ca_xxx classes, this
        # invokes __pos__ operator to return an instance of
        # the builtin base class
        return +value  # type: ignore


class DisconnectedCaConverter(CaConverter):
    def __getattribute__(self, __name: str) -> Any:
        raise NotImplementedError("No PV has been set as connect() has not been called")


class CaIntConverter(CaConverter[int]):
    def value(self, value: AugmentedValue) -> int:
        return int(value)  # type: ignore


class CaArrayConverter(CaConverter[np.ndarray]):
    def value(self, value: AugmentedValue) -> np.ndarray:
        # A less expensive conversion
        return np.array(value, copy=False)


class CaSequenceStrConverter(CaConverter[Sequence[str]]):
    def value(self, value: AugmentedValue) -> Sequence[str]:
        return [str(v) for v in value]  # type: ignore


class CaLongStrConverter(CaConverter[str]):
    def __init__(self):
        super().__init__(str, dbr.DBR_CHAR_STR, dbr.DBR_CHAR_STR)

    def write_value(self, value: Any) -> Any:
        # Add a null in here as this is what the commandline caput does
        # TODO: this should be in the server so check if it can be pushed to asyn
        return value + "\0"


class CaBoolConverter(CaConverter[bool]):
    def __init__(self):
        super().__init__(bool, dbr.DBR_SHORT)

    def value(self, value: AugmentedValue) -> bool:
        return bool(value)


class CaEnumConverter(CaConverter[str]):
    def __init__(self, supported_values: Mapping[str, str]):
        self.supported_values = supported_values
        super().__init__(
            str, dbr.DBR_STRING, metadata=SignalMetadata(choices=list(supported_values))
        )

    def value(self, value: AugmentedValue) -> str:
        return self.supported_values[str(value)]


_datatype_converter_from_dbr: dict[
    tuple[Dbr, bool], tuple[type[SignalDatatype], type[CaConverter]]
] = {
    (dbr.DBR_STRING, False): (str, CaConverter),
    (dbr.DBR_SHORT, False): (int, CaConverter),
    (dbr.DBR_FLOAT, False): (float, CaConverter),
    (dbr.DBR_ENUM, False): (str, CaConverter),
    (dbr.DBR_CHAR, False): (int, CaConverter),
    (dbr.DBR_LONG, False): (int, CaConverter),
    (dbr.DBR_DOUBLE, False): (float, CaConverter),
    (dbr.DBR_STRING, True): (Sequence[str], CaSequenceStrConverter),
    (dbr.DBR_SHORT, True): (Array1D[np.int16], CaArrayConverter),
    (dbr.DBR_FLOAT, True): (Array1D[np.float32], CaArrayConverter),
    (dbr.DBR_ENUM, True): (Sequence[str], CaSequenceStrConverter),
    (dbr.DBR_CHAR, True): (Array1D[np.uint8], CaArrayConverter),
    (dbr.DBR_LONG, True): (Array1D[np.int32], CaArrayConverter),
    (dbr.DBR_DOUBLE, True): (Array1D[np.float64], CaArrayConverter),
}


def make_converter(
    datatype: type | None, values: dict[str, AugmentedValue]
) -> CaConverter:
    pv = list(values)[0]
    pv_dbr = cast(
        Dbr, get_unique({k: v.datatype for k, v in values.items()}, "datatypes")
    )
    is_array = bool([v for v in values.values() if v.element_count > 1])
    # Make the datatype canonical for the inference below
    if datatype == typing.Sequence[str]:
        datatype = Sequence[str]
    # Infer a datatype and converter from the dbr
    inferred_datatype, converter_cls = _datatype_converter_from_dbr[(pv_dbr, is_array)]
    # Some override cases
    if is_array and pv_dbr == dbr.DBR_CHAR and datatype is str:
        # Override waveform of chars to be treated as string
        return CaLongStrConverter()
    elif not is_array and datatype is bool and pv_dbr == dbr.DBR_ENUM:
        # Database can't do bools, so are often representated as enums of len 2
        pv_num_choices = get_unique(
            {k: len(v.enums) for k, v in values.items()}, "number of choices"
        )
        if pv_num_choices != 2:
            raise TypeError(f"{pv} has {pv_num_choices} choices, can't map to bool")
        return CaBoolConverter()
    elif not is_array and pv_dbr == dbr.DBR_ENUM:
        pv_choices = get_unique(
            {k: tuple(v.enums) for k, v in values.items()}, "choices"
        )
        if enum_cls := get_enum_cls(datatype):
            # If explicitly requested then check
            return CaEnumConverter(get_supported_values(pv, enum_cls, pv_choices))
        elif datatype in (None, str):
            # Drop to string for safety, but retain choices as metadata
            return CaConverter(
                str,
                dbr.DBR_STRING,
                metadata=SignalMetadata(choices=list(pv_choices)),
            )
    elif (
        inferred_datatype is float
        and datatype is int
        and get_unique({k: v.precision for k, v in values.items()}, "precision") == 0
    ):
        # Allow int signals to represent float records when prec is 0
        return CaIntConverter(int, pv_dbr)
    elif datatype in (None, inferred_datatype):
        # If datatype matches what we are given then allow it and use inferred converter
        return converter_cls(inferred_datatype, pv_dbr)
    if pv_dbr == dbr.DBR_ENUM:
        inferred_datatype = "str | SubsetEnum | StrictEnum"
    raise TypeError(
        f"{pv} with inferred datatype {format_datatype(inferred_datatype)}"
        f" cannot be coerced to {format_datatype(datatype)}"
    )


# Cached call to avoid repeated initialization attempts
@cache
def _use_pyepics_context_if_imported():
    """Sets up the pyepics context if the module is imported."""
    ca = sys.modules.get("epics.ca", None)
    if ca:
        ca.use_initial_context()


class CaSignalBackend(EpicsSignalBackend[SignalDatatypeT]):
    """Backend for a signal to interact with PVs over channel access."""

    def __init__(
        self,
        datatype: type[SignalDatatypeT] | None,
        read_pv: str = "",
        write_pv: str = "",
    ):
        self.converter: CaConverter = DisconnectedCaConverter(float, dbr.DBR_DOUBLE)
        self.initial_values: dict[str, AugmentedValue] = {}
        self.subscription: Subscription | None = None
        self._all_updates = _all_updates()
        super().__init__(datatype, read_pv, write_pv)

    def source(self, name: str, read: bool):
        return f"ca://{self.read_pv if read else self.write_pv}"

    async def _store_initial_value(self, pv: str, timeout: float):
        try:
            self.initial_values[pv] = await caget(
                pv, format=FORMAT_CTRL, timeout=timeout
            )
        except CANothing as exc:
            logger.debug(f"signal ca://{pv} timed out")
            raise NotConnectedError(f"ca://{pv}") from exc

    async def connect(self, timeout: float):
        _use_pyepics_context_if_imported()
        if self.read_pv != self.write_pv:
            # Different, need to connect both
            await wait_for_connection(
                read_pv=self._store_initial_value(self.read_pv, timeout=timeout),
                write_pv=self._store_initial_value(self.write_pv, timeout=timeout),
            )
        else:
            # The same, so only need to connect one
            await self._store_initial_value(self.read_pv, timeout=timeout)
        self.converter = make_converter(self.datatype, self.initial_values)

    async def _caget(self, pv: str, format: Format) -> AugmentedValue:
        return await caget(
            pv, datatype=self.converter.read_dbr, format=format, timeout=None
        )

    def _make_reading(self, value: AugmentedValue) -> Reading[SignalDatatypeT]:
        return {
            "value": self.converter.value(value),
            "timestamp": value.timestamp,
            "alarm_severity": -1 if value.severity > 2 else value.severity,
        }

    async def put(self, value: SignalDatatypeT | None, wait: bool):
        if value is None:
            write_value = self.initial_values[self.write_pv]
        else:
            write_value = self.converter.write_value(value)
        try:
            await caput(
                self.write_pv,
                write_value,
                datatype=self.converter.write_dbr,
                wait=wait,
                timeout=None,
            )
        except CANothing as exc:
            # If we ran into a write error, check to see if there is a list
            # of valid choices, and if the value we tried to write is in that list.
            valid_choices = self.converter.metadata.get("choices")
            if valid_choices:
                if value not in valid_choices:
                    msg = (
                        f"{value} is not a valid choice for {self.write_pv}, "
                        f"valid choices: {self.converter.metadata.get('choices')}"
                    )
                    raise ValueError(msg) from exc
                raise
            raise

    async def get_datakey(self, source: str) -> DataKey:
        value = await self._caget(self.read_pv, FORMAT_CTRL)
        metadata = _metadata_from_augmented_value(
            self.datatype, value, self.converter.metadata
        )
        return make_datakey(
            self.converter.datatype, self.converter.value(value), source, metadata
        )

    async def get_reading(self) -> Reading[SignalDatatypeT]:
        value = await self._caget(self.read_pv, FORMAT_TIME)
        return self._make_reading(value)

    async def get_value(self) -> SignalDatatypeT:
        value = await self._caget(self.read_pv, FORMAT_RAW)
        return self.converter.value(value)

    async def get_setpoint(self) -> SignalDatatypeT:
        value = await self._caget(self.write_pv, FORMAT_RAW)
        return self.converter.value(value)

    def set_callback(self, callback: Callback[Reading[SignalDatatypeT]] | None) -> None:
        if callback and self.subscription:
            msg = "Cannot set a callback when one is already set"
            raise RuntimeError(msg)

        if self.subscription:
            self.subscription.close()
            self.subscription = None

        if callback:
            self.subscription = camonitor(
                self.read_pv,
                lambda v: callback(self._make_reading(v)),
                datatype=self.converter.read_dbr,
                format=FORMAT_TIME,
                all_updates=self._all_updates,
            )
